/* ============================= C MeatAxe ==================================
   File:        $Id $
   Comment:     Input/output of integers.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include "config.h"
#include <stdint.h>

MTX_DEFINE_FILE_INFO

/**
 ** @addtogroup ff2
 ** @{
 **/


/**
 ** Read long integers.
 ** This function reads @ n long integers from the file @a f into the array 
 ** @a buf. @a buf must point to a memory area of at least n*sizeof(long) 
 ** bytes and the file must be open for reading. The return value indicates how
 ** many integers have actually been read. This number may be less than
 ** @a n because the end of file was encountered while reading. A negative
 ** return value indicates a file i/o error.
 ** 
 ** %SysReadLong32() expects that the numbers in the file are 4-byte integers 
 ** in little-endian format, i.e. the least significant byte first.
 ** Using a machine-independent data format makes MeatAxe data files 
 ** more portable, but there are also some disadvantages:
 ** - The conversion to and from machine-independent format involves several
 **   arithmetic operations for each number read/written.
 ** - The highest number which can be read/written is 2<sup>32</sup>-1.
 ** @param f File to read from.
 ** @param buf Pointer to buffer.
 ** @param n Number of integers to read.
 ** @return Number of integers that were actually read, or $/1$ on error.
 **/

int SysReadLong32(FILE *f, long *buf, int n)
{
    int nread;

    /* Check the arguments
       ------------------- */
    if (f == NULL || buf == NULL || n < 0)
    {
	MTX_ERROR3("Invalid arguments (f:%s,buf:%s,n=%d)",f ? "ok" : "NULL",
		buf ? "ok" : "NULL",n);
	return -1;
    }

    /* Read <n> long integers
       ---------------------- */
#if MTX_CONFIG_BIG_ENDIAN
    for (nread = 0; nread < n; ++nread)
    {
	uint32_t a = 0;
	if (fread(&a,1,sizeof(a),f) != 4) break;
	*buf = (long)(int32_t)__builtin_bswap32(a);
	++buf;
    }
#elif MTX_CONFIG_LONG64
    for (nread = 0; nread < n; ++nread)
    {
	int32_t a = 0;
	if (fread(&a,1,sizeof(int32_t),f) != 4) break;
	*buf = (long)a;
	++buf;
    }
#else
    nread = (int)fread((char *)buf,sizeof(long),(size_t)n,f);
    /* printf("nread=%d, error=%d, eof=%d, GetLastError()=%d\n",
       nread,ferror(f),feof(f),GetLastError());
     */
#endif
    /* Check for errors
       ---------------- */
    if (ferror(f) && !feof(f))
    {
	MTX_ERROR("Read failed: %S");
	return -1;
    }
    return nread;
}



/**
 ** Write long integers.
 ** This function writes @a n long integers from the the array @a buf to the 
 ** file @a f. @a buf must point to a memory area of at least n*sizeof(long) 
 ** bytes and @a f must be open for writing. The numbers are written in a 
 ** machine-independent format which can be read by SysReadLong().
 ** @param f File to write to.
 ** @param buf Pointer to buffer.
 ** @param n Number of integers to write.
 ** @return Number of integers that were written, or $-1$ on error.
 **/

int SysWriteLong32(FILE *f, const long *buf, int n)
{
    unsigned char a[4];
    int nwritten;

#if MTX_CONFIG_BIG_ENDIAN
    for (nwritten = 0; nwritten < n; ++nwritten)
    {
	uint32_t a = __builtin_bswap32((uint32_t)(int32_t)*buf);
	if (fwrite(&a,1,sizeof(a),f) != 4) break;
	++buf;
    }
#elif MTX_CONFIG_LONG64
    for (nwritten = 0; nwritten < n; ++nwritten)
    {
	int32_t a = (int32_t)*buf;
	if (fwrite(&a,1,sizeof(a),f) != 4) break;
	++buf;
    }
#else
    nwritten = fwrite((char *)buf,sizeof(long),n,f);
#endif
    return nwritten;
}


int SysReadLongX(FILE *f, long *buf, int n_bytes)
{
    int n_read;			/* Bytes read so far */
#if MTX_CONFIG_BIG_ENDIAN
    size_t to_read;
    for (n_read = 0; n_read < n_bytes; n_read += to_read)
    {
	unsigned long a = 0;
	if (n_bytes - n_read < sizeof(long))
	{
	    to_read = n_bytes - n_read;
	}
	else
	{
	    to_read = sizeof(long);
	}
	to_read = fread(&a,1,to_read,f);
	if (to_read == 0) break;
#if MTX_CONFIG_LONG32
	*buf = (long)__builtin_bswap32(a);
#else
	*buf = (long)__builtin_bswap64(a);
#endif
	++buf;
    }
#else
    n_read = fread(buf,1,n_bytes,f);
#endif
    return n_read;
}


int SysWriteLongX(FILE *f, const long *buf, int n_bytes)
{
    int n_written;
#if MTX_CONFIG_BIG_ENDIAN
    size_t to_write;
    for (n_written = 0; n_written < n_bytes; n_written += to_write)
    {
	if (n_bytes - n_written < sizeof(long))
	{
	    to_write = n_bytes - n_written;
	}
	else
	{
	    to_write = sizeof(long);
	}
#if MTX_CONFIG_LONG32
	unsigned long a = __builtin_bswap32((unsigned long)*buf);
#else
	unsigned long a = __builtin_bswap64((unsigned long)*buf);
#endif
	if (fwrite(&a,1,to_write,f) != to_write) break;
	++buf;
    }
#else
    n_written = fwrite(buf,1,n_bytes,f);
#endif
    return n_written;
}



/**
 ** @}
 **/

